//
// Copyright (c) 2016 Samsung Electronics Co., Ltd.
//
// Licensed under the Apache License, Version 2.0 (the License);
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#include <app_manager.h>
#include <dlog.h>
#include <errno.h>
#include <glib.h>
#include <libxml/parser.h>
#include <libxml/tree.h>
#include <pkgmgr-info.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <tzplatform_config.h>

/* Define EXPORT_API */
#ifndef EXPORT_API
#define EXPORT_API __attribute__((visibility("default")))
#endif

#ifdef LOG_TAG
#undef LOG_TAG
#endif
#define LOG_TAG "tts-engine-parser"

#define TTS_TAG_ENGINE_LANG_BASE		"tts-engine-language-pack"
#define TTS_TAG_ENGINE_NAME			"name"
#define TTS_TAG_ENGINE_ID			"id"
#define TTS_TAG_ENGINE_VOICE_SET		"voices"
#define TTS_TAG_ENGINE_VOICE			"voice"
#define TTS_TAG_ENGINE_VOICE_TYPE		"type"

#define TTS_CONFIG_BASE		tzplatform_mkpath(TZ_USER_HOME, "share/.voice")
#define TTS_HOME		tzplatform_mkpath(TZ_USER_HOME, "share/.voice/tts")
#define TTS_ENGINE_BASE		tzplatform_mkpath(TZ_USER_HOME, "share/.voice/tts/1.0")
#define TTS_ENGINE_INFO		tzplatform_mkpath(TZ_USER_SHARE, ".voice/tts/1.0/engine-info")

#define TTS_METADATA_LANGUAGE	"http://tizen.org/metadata/tts-engine-language-pack/language"

typedef struct metadata {
	const char *key;
	const char *value;
} metadata;

static xmlDocPtr g_doc;

static int __create_engine_info_xml(const char *pkgid)
{
	LOGD("=== Create engine info doc");
	g_doc = xmlNewDoc((xmlChar*)"1.0");
	if (NULL == g_doc) {
		LOGE("[ERROR] Fail to new doc");
		return -1;
	}
	LOGD("===");
	return 0;
}

static int __save_engine_info_xml(const char *pkgid)
{
	LOGD("=== Save engine info doc");

	/* Make directories */
	if (0 != access(TTS_CONFIG_BASE, F_OK)) {
		if (0 != mkdir(TTS_CONFIG_BASE, S_IRUSR | S_IWUSR | S_IXUSR | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH)) {
			LOGE("[ERROR] Fail to make directory : %s", TTS_CONFIG_BASE);
			return -1;
		} else {
			LOGD("Success to make directory : %s", TTS_CONFIG_BASE);
		}
	}

	if (0 != access(TTS_HOME, F_OK)) {
		if (0 != mkdir(TTS_HOME, S_IRUSR | S_IWUSR | S_IXUSR | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH)) {
			LOGE("[ERROR] Fail to make directory : %s", TTS_HOME);
			return -1;
		} else {
			LOGD("Success to make directory : %s", TTS_HOME);
		}
	}

	if (0 != access(TTS_ENGINE_BASE, F_OK)) {
		if (0 != mkdir(TTS_ENGINE_BASE, S_IRUSR | S_IWUSR | S_IXUSR | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH)) {
			LOGE("[ERROR] Fail to make directory : %s", TTS_ENGINE_BASE);
			return -1;
		} else {
			LOGD("Success to make directory : %s", TTS_ENGINE_BASE);
		}
	}

	if (0 != access(TTS_ENGINE_INFO, F_OK)) {
		if (0 != mkdir(TTS_ENGINE_INFO, S_IRUSR | S_IWUSR | S_IXUSR | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH)) {
			LOGE("[ERROR] Fail to make directory : %s", TTS_ENGINE_INFO);
			return -1;
		} else {
			LOGD("Success to make directory : %s", TTS_ENGINE_INFO);
		}
	}

	char path[256] = {'\0',};
	snprintf(path, 256, "%s/%s.xml", TTS_ENGINE_INFO, pkgid);
	int ret = xmlSaveFormatFile(path, g_doc, 1);
	LOGD("xmlSaveFile (%d)", ret);
	LOGD("===");
	return 0;
}

static int __remove_engine_info_xml(const char *pkgid)
{
	LOGD("=== Remove engine info doc");
	char path[256] = {'\0',};
	snprintf(path, 256, "%s/%s.xml", TTS_ENGINE_INFO, pkgid);
	if (0 == access(path, F_OK)) {
		LOGD("Remove engine info xml(%s)", path);
		remove(path);
	}
	LOGD("===");
	return 0;
}

static void __insert_language_from_metadata(xmlNodePtr root, const char *language)
{
	LOGD("==== Insert language");
	char* voice = NULL;
	char* lang = NULL;
	char* type = NULL;

	char *tmp_lang = NULL;
	char *tmp_free = NULL;
	tmp_free = tmp_lang = strdup(language);
	if (NULL == tmp_lang) {
		LOGE("Fail to memory allocation");
		return;
	}
	xmlNodePtr voices_node = NULL;
	xmlNodePtr voice_node = NULL;

	voices_node = xmlNewNode(NULL, (const xmlChar*)TTS_TAG_ENGINE_VOICE_SET);

	voice = strsep(&tmp_lang, ",");
	while (NULL != voice) {
		LOGD("voice (%s)", voice);
		voice_node = xmlNewNode(NULL, (const xmlChar*)TTS_TAG_ENGINE_VOICE);
		lang = strsep(&voice, ":");
		LOGD("lang (%s)", lang);
		type = strsep(&voice, " ");
		LOGD("type (%s)", type);
		xmlSetProp(voice_node, (const xmlChar*)TTS_TAG_ENGINE_VOICE_TYPE, (const xmlChar*)type);
		xmlNodeSetContent(voice_node, (const xmlChar*)lang);
		xmlAddChild(voices_node, voice_node);
		voice = strsep(&tmp_lang, ",");
	}
	xmlAddChild(root, voices_node);

	free(tmp_free);
	tmp_free = NULL;
}

EXPORT_API
int PKGMGR_MDPARSER_PLUGIN_INSTALL(const char *pkgid, const char *appid, GList *list)
{
	LOGD("METADATA INSTALL");
	LOGD("pkgid(%s) appid(%s) list(%d)", pkgid, appid, g_list_length(list));

	if (0 >= g_list_length(list)) {
		LOGE("[ERROR] No Engine Metadata");
		return 0;
	}

	GList *iter = NULL;
	metadata *md = NULL;

	__create_engine_info_xml(pkgid);

	xmlNodePtr root = NULL;
	xmlNodePtr cur = NULL;

	root = xmlNewNode(NULL, (const xmlChar*)TTS_TAG_ENGINE_LANG_BASE);
	if (NULL == root) {
		LOGE("[ERROR] Fail to get new node");
		xmlFreeDoc(g_doc);
		return -1;
	}
	xmlDocSetRootElement(g_doc, root);

	/* Save name */
	cur = xmlNewNode(NULL, (const xmlChar*)TTS_TAG_ENGINE_NAME);
	xmlNodeSetContent(cur, (const xmlChar*)pkgid);
	xmlAddChild(root, cur);

	iter = g_list_first(list);
	while (NULL != iter) {
		md = (metadata *)iter->data;
		if (NULL != md && NULL != md->key) {
			LOGD(" - key(%s) value(%s)", md->key, md->value);
			if (!strcmp(md->key, TTS_METADATA_LANGUAGE)) {
				__insert_language_from_metadata(root, md->value);
			} else {
				LOGW("[WARNING] Unknown metadata type");
			}
		}
		iter = g_list_next(iter);
	}

	LOGD("");

	__save_engine_info_xml(pkgid);
	xmlFreeDoc(g_doc);

	return 0;
}

EXPORT_API
int PKGMGR_MDPARSER_PLUGIN_UNINSTALL(const char *pkgid, const char *appid, GList *list)
{
	LOGD("METADATA UNINSTALL");
	LOGD("pkgid(%s) appid(%s) list(%d)", pkgid, appid, g_list_length(list));

	GList *iter = NULL;
	metadata *md = NULL;

	iter = g_list_first(list);
	while (NULL != iter) {
		md = (metadata *)iter->data;
		if (NULL != md && NULL != md->key) {
			LOGD(" - key(%s) value(%s)", md->key, md->value);
		}
		iter = g_list_next(iter);
	}

	__remove_engine_info_xml(pkgid);

	LOGD("");
	return 0;
}

EXPORT_API
int PKGMGR_MDPARSER_PLUGIN_UPGRADE(const char *pkgid, const char *appid, GList *list)
{
	LOGD("METADATA UPGRADE");
	LOGD("pkgid(%s) appid(%s) list(%d)", pkgid, appid, g_list_length(list));

	PKGMGR_MDPARSER_PLUGIN_UNINSTALL(pkgid, appid, list);
	PKGMGR_MDPARSER_PLUGIN_INSTALL(pkgid, appid, list);

	LOGD("");
	return 0;
}
